%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%  File: eyeFinderHandout.m
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This Matlab code provides a demonstration of skin-coloured blob
% tracking followded by a PCA based eye finder.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Prerequisites: 
%%%  1. The singular value decomposition.  Eg. see the
%%%     svdTutorial from the Standford iseToolbox.
%%%  2. Colour selection, see pepperDemo.m.
%%%  3. Image filtering, see cannyTutorial.m
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%  Dependencies:
%       teaching/iseToolbox/utility/
%          warpAffine2.m
%       teaching/iseToolbox/pyrTools/
%          corrDn.m showIm.m
%       teaching/iseToolbox/filters/
%          rconv2sep.m 
%       teaching/utvisToolbox/file
%          resizeImageFig.m
%          ppmRead.m
%       ./ppm/    
%         frame0###.ppm, for frame numbers = 10:50:660
%
%%%  Author: ADJ, Fall 2002
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear;
close all;

%% CHANGE these to the directories the ISE toolbox, the UTvis toolbox
%% and the images are stored.
global matlabVisRoot
if length(matlabVisRoot)==0
  startup;
end

iseRoot = [matlabVisRoot '/iseToolbox'];
utvisRoot = [matlabVisRoot '/utvisToolbox'];

TRUE = (1 == 1);
FALSE = ~TRUE;

DEBUG = TRUE;
SHOW_TRACKING = TRUE;
EYE_FINDER = FALSE;

%% Set the path to include these directories
path(path, [utvisRoot '/file']);
path(path, [iseRoot '/pyrTools/MEX']);
path(path, [iseRoot '/pyrTools']);
path(path, [iseRoot '/filters']);
path(path, [iseRoot '/utility']);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Read saved information about eigen eyes
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
path(path, [utvisRoot '/tutorials/eigenTut']);
load 'trainAndTestEyes.mat' 

%% Load saved information from eigenEyes.
sizeEyeIm = sizeIm;
dcEyeIm = ones(sizeEyeIm)/sqrt(prod(sizeEyeIm));
dcEyeIm = dcEyeIm(:);

%% Set the number of basis elements to use in fitting eyes
nBasis= 5;

%% Show out of subspace std dev for each pixel, as in question 2.
varIm=basisIm(:,(nBasis+1):end).^2 * ...
      sigmaUniEye((nBasis+1):end).^2;
if DEBUG
  figure(10); clf; showIm(reshape(sqrt(varIm), sizeEyeIm)); 
  title(sprintf('Out of subspace Stdev for nBasis = %d',nBasis));
  fprintf(2,'Press any key to continue...');
  pause;
  fprintf(2,'ok\n');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Frames from the colour image sequence
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Read and display the frames to be used in this demonstration.
imRoot = './ppm/';
fRoot = 'fDec';

% Loop over the image filenames in fn.
for frameNum = 10:50:660

  % Read the kIm^th image
  im = ppmRead([imRoot fRoot num2strPad(frameNum, 4) '.ppm']);
  im = im/255;
  
  h = figure(1); clf; image(im);
  resizeImageFig(h, size(im), 1);
  drawnow;
  pause(0.1);
  
end
sizeIm = [240 320];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Skin Colour Training
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% For one of the colour images, select a small rectangular
% region primarily containing skin tones.
figure(1);
fprintf(2,'Mouse in top left then bottom right corners\n');
fprintf(2,'of a skin coloured rectangle in Figure 1...\n');
pts = ginput(2);
bbox = [floor(pts(1,:)); ceil(pts(2,:))];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Question 2a goes HERE.

%% End Question 2a.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Initialization code for Questions 2 and 3 goes HERE.

%% End of new initialization code.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% These simple coordinate images are useful for finding
%% the mean location of a response image...
[xMesh yMesh] = meshgrid(1:sizeIm(2), 1:sizeIm(1));
if DEBUG
  figure(10); clf; 
  subplot(1,2,1); showIm(xMesh); title('x coordinate');
  subplot(1,2,2); showIm(yMesh); title('y coordinate');
  fprintf(2,'Press any key to continue...');
  pause;
  fprintf(2,'ok\n');
end

%% Initial guess for center location for colour blob.
%% You should not need to change this.
meanLoc = [sizeIm(2) sizeIm(1)]/2; %% i.e. [x_loc y_loc]

% Loop over the images
for frameNum = 10:50:660

  %% Read and display current image
  fname = [imRoot fRoot num2strPad(frameNum, 4) '.ppm'];
  fprintf(1,'%s\n', fname);
  im = ppmRead(fname);
  sizeIm = size(im);
  sizeIm = sizeIm(1:2);
  
  h = figure(1); clf; image(reshape(im/255, [sizeIm 3]));
  resizeImageFig(h, sizeIm, 1);
  pause(0.1);
  
  %% Build skin colour response image
  fprintf(2,'Detecting skin coloured blobs...');
  tic;
  rgbIm = reshape(im, prod(sizeIm), 3);
  skinIm = zeros(sizeIm);
  nrmIm = zeros(sizeIm);
  blurSkinIm = zeros(sizeIm);
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %% Question 2b goes HERE.
  %% Identify skin coloured pixels
  
  
  %% Question 2c goes HERE.
  %% Blur skin coloured pixels and threshold response.


  %% Keep code below for displaying results of question 3b, and 3c.
  t = toc;
  fprintf(2,sprintf('done (%f secs)\n',t));
  
  %% Display skin colour results
  figure(2); clf; subplot(1,2,1);
  showIm(skinIm);
  title('Skin coloured pixels (white)');
 
  figure(2);
  subplot(1,2,2);
  showIm(blurSkinIm +  0.25*im(:,:,1)/255);
  title('Skin colour blobs (white)');
  pause(0.1);
  %% End Questions 2b,c.
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %%% Colour Blob Tracking
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %% Given the mean location for the previous frame (or the 
  %% initial guess), find the location of a large isolated blob
  %% in the current frame using the blurSkinIm
  fprintf(2,'Tracking skin coloured blob...');
  tic;
  wghtIm = zeros(sizeIm);
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %% Question 2d goes HERE.

  %% End Question 2d.
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  t = toc;
  fprintf(2,sprintf('done (%f secs)\n',t));
  
  %% meanLoc should now provide an image coordinate close to the
  %% blob in blurSkinIm associated with the face
  if DEBUG
    figure(10); clf;
    showIm(wghtIm + 0.2*blurSkinIm);
    title('Determining mean position of blob');
    fprintf(2,'Press any key to continue...');
    pause;
    fprintf(2,'ok\n');
  end
  
  fprintf(2,'Warping image...');
  tic;
  %% Crop and rescale the image around the face
  scl = 0.65;  %% 1/scl gives the expansion factor to get
               %% the inter-eye distance to be about 40 pixels.
               %% This was the scale used for the eigen-eyes.
  %% Do an affine warp of the image, expanding by 1/scl and
  %% translating so that meanLoc is at pixel (x,y) = (160, 120).
  im = reshape(im, [sizeIm 3]);
  A = [diag([scl scl]) (1-scl)*(meanLoc')+scl*(meanLoc' - [160;120]) ];
  wIm=warpAffine2(im(:,:,1), A);
  %% Crop the warped image to the original image size, 320 x 240.
  wIm = wIm(1:240,1:320);
  %% Set any pixels mapped from outside the original image to 0.
  if any(isnan(wIm(:)))
      wIm(isnan(wIm)) = 0;
  end

  %% Crop and rescale the skin blob image.
  blurSkinIm=warpAffine2(blurSkinIm, A);
  t = toc;
  fprintf(2,'done (%f secs)\n',t);
  
  if SHOW_TRACKING
    %% Show rescaled image regions
    figure(10); clf;
    showIm(blurSkinIm + 0.2 * wIm/255);
    title('Recentered and rescaled image');
    if DEBUG
      fprintf(2,'Press any key to continue...');
      pause;
      fprintf(2,'ok\n');
    else
      pause(0.1);  % Forces raise and display window.
    end
  end
  
  if EYE_FINDER
    fprintf(2,'Looking for eyes...');
    tic;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Question 3 goes HERE.
    rangeEyeY = (-14):10;
    rangeEyeX = (-9):10;

    %% End Question 3.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    t = toc;
    fprintf(2,sprintf('done (%f secs)\n',t));
    
    fprintf(2,'Press any key to continue...');
    pause;
    fprintf(2,'ok\n');
  end  % Eye finder

end % loop over frames


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
